package com.hisschemoller.audio.model 
{
	import de.popforge.parameter.MappingBoolean;
	import de.popforge.parameter.MappingIntLinear;
	import de.popforge.parameter.MappingNumberLinear;

	import com.hisschemoller.audio.model.enum.AudioProcessorType;
	import com.hisschemoller.audio.model.parameter.Parameter;
	import com.hisschemoller.audio.model.processor.IAudioInstrument;
	import com.hisschemoller.audio.model.processor.IAudioProcessor;
	import com.hisschemoller.audio.model.processor.PitchedSampleProcessor;
	import com.hisschemoller.audio.model.vo.PitchedSampleProgramVO;
	import com.hisschemoller.audio.model.vo.ProgramVO;
	import com.hisschemoller.audio.model.vo.SoundVO;
	import com.hisschemoller.audio.notification.AudioNotifications;
	import com.hisschemoller.audio.utils.Conversions;
	import com.hisschemoller.ui.model.vo.AlertVO;
	import com.hisschemoller.ui.notification.UINotifications;
	import com.hisschemoller.utils.midi.MidiSpecification;
	import com.newgonzo.midi.messages.ChannelMessage;
	import com.newgonzo.midi.messages.VoiceMessage;

	import org.puremvc.as3.multicore.interfaces.IProxy;

	/**
	 * @author wouter.hisschemoller
	 */
	public class SamplerInstrumentProxy extends BaseInstrumentProxy implements IProxy, IAudioInstrument
	{
		public static const PROGRAM_START_CHANGED : String = "programStartChanged";
		public static const PROGRAM_END_CHANGED : String = "programEndChanged";
		public static const PROGRAM_TRANSPOSE_CHANGED : String = "programTransposeChanged";
		public static const PROGRAM_TUNE_CHANGED : String = "programTuneChanged";
		public static const PROGRAM_REVERSE_CHANGED : String = "programReverseChanged";
		private var _processors : Vector.<PitchedSampleProcessor> = new Vector.<PitchedSampleProcessor>( );
		private var _programs : Vector.<PitchedSampleProgramVO> = new Vector.<PitchedSampleProgramVO>( MidiSpecification.MAX_VALUE_7BIT, true );
		private var _soundID : String;
		private var _soundVO : SoundVO;
		private var _transpose : Parameter = new Parameter( new MappingIntLinear( -12, 12 ), 0 );
		private var _tune : Parameter = new Parameter( new MappingNumberLinear( -0.5, 0.5 ), 0 );
		private var _reverse : Parameter = new Parameter( new MappingBoolean( ), 0 );
		private var _start : int;
		private var _end : int;

		public function SamplerInstrumentProxy (name : String)
		{
			super( name );
		}

		/**
		 * Parse program XML, create programVO and add to _programs.
		 * @param programXML: XML program definition.
		 */
		public function addProgram ( programXML : XML ) : void
		{
			/** Check for the right program type. */
			if(programXML.@type != AudioProcessorType.PITCHED_SAMPLE_PROCESSOR)
			{
				return;
			}
			
			/** Note: Vector.<ProgramVO>(_programs) downcasts the vector type. */
			var programIndex : uint = getValidProgramIndex( Vector.<ProgramVO>( _programs ), programXML.@index );
			if(programIndex > -1)
			{
				var programVO : PitchedSampleProgramVO = new PitchedSampleProgramVO( );
				programVO.parseXML( programXML );
				
				_programs[programIndex] = programVO;
				
				log( "n Program \"" + programVO.id + "\" added to instrument " + getProxyName( ) + "." );
			}
			else
			{
				facade.sendNotification( UINotifications.ALERT, new AlertVO( "Program not loaded", "Program " + programVO.id + " couldn't be loaded in instrument " + getProxyName( ) + " because all 128 program slots are in use." ) );
			}
		}

		/**
		 * Save an audio file using the current settings.
		 */
		//		public function saveAudioFile() : void
		//		{
		//			/** If there's no program there's nothing to save. */
		//			if(_selectedProgramIndex == -1)
		//			{
		//				facade.sendNotification(UINotification.ALERT, new AlertVO("Unable to save", "There's no program to save as an audio file."));
		//				return;
		//			}
		//			
		//			/** The number of samples needed with transpose, tune, start and end applied. */
		//			var rate : Number = Conversions.semitoneToRate(_transpose.value + _tune.value);
		//			var numSamples : uint = (_end - _start) / rate;
		//			
		//			/** Create vector to hold the samples. */
		//			var samples : Vector.<Vector.<Number>> = new Vector.<Vector.<Number>>(2, true);
		//			samples[0] = new Vector.<Number>(numSamples, true);
		//			samples[1] = new Vector.<Number>(numSamples, true);
		//			
		//			/** Let a processor fill the vector with sample data. */
		//			var processor : PitchedSampleProcessor = new PitchedSampleProcessor(_soundVO.byteArray, _start, _end);
		//			processor.rate = rate;
		//			processor.process(samples, 0, samples[0].length);
		//			
		//			facade.sendNotification(AppFacade.FILE_SAVE, samples);
		//		}

		/**
		 * Let each processor add it's data to the buffer.
		 * @param buffer : stereo buffer of AudioDriver.BUFFER_SIZE number of samples
		 * @param bufferStartIndex: From where in the buffer to start to write.
		 */
		public function process (buffer : Vector.<Vector.<Number>>, bufferStartIndex : uint, bufferLength : uint) : void
		{
			var n : int = _processors.length;
			while(--n > -1)
			{
				/** IAudioProcessor returns true if the sound is finished. */
				if(_processors[n].process( buffer, bufferStartIndex, AudioDriverProxy.BUFFER_BLOCK_SIZE ))
				{
					facade.sendNotification( AudioNotifications.PROCESSOR_REMOVED, _processors[n], getProxyName( ) );
					
					/** Remove the event from the _processors list. */
					_processors.splice( n, 1 );
				}
			}
		}

		/**
		 * Play a sound.
		 */
		public function noteOn (message : VoiceMessage, deltaTime : int = 0) : void
		{
			if(_isReadyToPlay)
			{
				var processor : PitchedSampleProcessor = new PitchedSampleProcessor( _soundVO.samples, _start, _end, deltaTime, message.pitch, message.velocity );
				processor.rate = Conversions.semitoneToRate( _transpose.value + _tune.value );
				
				addProcessor( processor );
			}
		}

		/**
		 * Stop a sound.
		 */
		public function noteOff (message : VoiceMessage, deltaTime : int = 0) : void
		{
			var n : int = _processors.length;
			while(--n > -1)
			{
				if(_processors[n].pitch == message.pitch) _processors[n].stop( );
			}
		}

		/**
		 * Stop all sounds.
		 */
		public function allNotesOff () : void
		{
			var n : int = _processors.length;
			while(--n > -1)
			{
				_processors[n].stop( );
			}
		}

		/**
		 * @param message: ChannelMessage.
		 * @param deltaTime: Time offset in samples. Not used at the moment.
		 */
		public function programChange (message : ChannelMessage, deltaTime : int = 0) : void 
		{
			if(message.data1 > -1)
			{
				if(_programs[message.data1])
				{
					_selectedProgramIndex = message.data1;
					
					var programVO : PitchedSampleProgramVO = _programs[message.data1];
					_start = programVO.start;
					_end = programVO.end;
					_transpose.value = programVO.transpose;
					_tune.value = programVO.tune;
					_reverse.value = programVO.reverse;
					_soundID = programVO.soundID;
					
					/** If no end sample index was provided set it to the total sample length.  */
					if( _end == 0 )
					{
						var audioLibrary : AudioLibraryProxy = facade.retrieveProxy( AudioLibraryProxy.NAME ) as AudioLibraryProxy;
						_end = audioLibrary.getReferenceToSoundByID( programVO.soundID ).numSamplesTotal;
					}
					
					/** 
					 * NOTE (for instruments that play samples): When the
					 * program changes the instrument gets a COPY of the SoundVO
					 * object from the AudioLibraryProxy. This allows for the 
					 * instrument to perform destructive editing on the sound. 
					 * Reversing it's sound for example.
					 */
					
					/** If no sound exists or it does exist but it's not the required sound. */
					if(!_soundVO || (_soundVO && (_soundVO.id != _soundID)))
					{
						_soundVO = (facade.retrieveProxy( AudioLibraryProxy.NAME ) as AudioLibraryProxy).getCopyOfSoundByID( _soundID );
					}
					
					/** If the SoundVO exists the sampler can play it's sound. */
					if(_soundVO)
					{
						_isReadyToPlay = true;
						
						if(_end == -1)
						{
							_end = _soundVO.numSamplesTotal;
						}
					}
					else
					{
						_isReadyToPlay = false;
					}
				}
				else
				{
					/** Program doesn't exist. Change nothing. */
					return;
				}
			}
			else
			{
				/** Program number = -1, clear all settings. */
				_selectedProgramIndex = -1;
				
				_start = -1;
				_end = -1;
				_transpose.value = 0;
				_tune.value = 0;
				_reverse.value = false;
				_soundID = "";
				_soundVO = null;
				
				_isReadyToPlay = false;
			}

			facade.sendNotification( PROGRAM_START_CHANGED, _start );
			facade.sendNotification( PROGRAM_END_CHANGED, _end );
			facade.sendNotification( PROGRAM_TRANSPOSE_CHANGED, _transpose.value );
			facade.sendNotification( PROGRAM_TUNE_CHANGED, _tune.value );
			facade.sendNotification( PROGRAM_REVERSE_CHANGED, _reverse.value );
		}

		/**
		 * Adjust start position of the sample (for playback).
		 * @param startIndex : Index of start sample within audio file.
		 */
		public function set programStart (startIndex : int) : void
		{
			_start = startIndex;
			
			facade.sendNotification( PROGRAM_START_CHANGED, _start );
		}

		/**
		 * Adjust end position of the sample (for playback).
		 * @param endIndex : Index of end sample within audio file.
		 */
		public function set programEnd (endIndex : int) : void
		{
			_end = endIndex;
			
			facade.sendNotification( PROGRAM_END_CHANGED, _end );
		}

		/**
		 * Change transpose.
		 * @param transpose : transpose value of the sample. Usually in semitones.
		 */
		public function set programTranspose (transpose : Number) : void
		{
			_transpose.value = transpose;
			
			updateRateOnAllProcessors( );
			
			facade.sendNotification( PROGRAM_TRANSPOSE_CHANGED, _transpose.value );
		}

		/**
		 * Change tune.
		 * @param tune : tune value of the sample. Usually up to half a semitone up or down.
		 */
		public function set programTune (tune : Number) : void
		{
			_tune.value = tune;
			
			updateRateOnAllProcessors( );
			
			facade.sendNotification( PROGRAM_TUNE_CHANGED, _tune.value );
		}

		/**
		 * Adjust start position of the sample (for playback).
		 * @param reverse : transpose value of the sample. Usually in semitones.
		 */
		public function set programReverse (reverse : Boolean) : void
		{
			_reverse.value = reverse;
			
			/** Stop the sound that currently plays (if there is any). */
			allNotesOff( );
			
			/** The actual reverse operation is done by ReverseSampleCommand. */
			// facade.sendNotification(AudioNotifications.REVERSE_SAMPLE, _soundVO);
			
			/** Adjust the start and end point. */
			var newEnd : int = _soundVO.numSamplesTotal - _start;
			programStart = _soundVO.numSamplesTotal - _end;
			programEnd = newEnd;
			
			/** This should be done by the display actually. */
			// var displayProxy : WaveDisplayProxy = WaveDisplayProxy(facade.retrieveProxy(getProxyName() + WaveDisplayProxy.NAME_SUFFIX));
			// displayProxy.firstSampleInView = _soundVO.numSamplesTotal - displayProxy.firstSampleInView - displayProxy.numSamplesInView;
			// displayProxy.forceUpdate();

			facade.sendNotification( PROGRAM_REVERSE_CHANGED, _reverse.value );
		}

		/**
		 * @param processor : IAudioProcessor
		 */
		private function addProcessor (processor : IAudioProcessor) : void
		{
			if(!processor is PitchedSampleProcessor) return;
			
			_processors.splice( 0, 0, processor );
			
			facade.sendNotification( AudioNotifications.PROCESSOR_ADDED, processor, getProxyName( ) );
		}

		/**
		 * When transpose or tune are changed, update the playback speed of all processors.
		 */
		private function updateRateOnAllProcessors () : void
		{
			var rate : Number = Conversions.semitoneToRate( _transpose.value + _tune.value );
			var n : int = _processors.length;
			while(--n > -1)
			{
				_processors[n].rate = rate;
			}
		}
	}
}
