package com.hisschemoller.audio.model.processor 
{

	/**
	 * Plays part of an audio file from _sourceStartIndex to _sourceEndIndex.
	 * Playback speed can be changed by adjusting _rate.
	 * 
	 * @author Wouter Hisschemoller
	 * (c) Apr 7, 2010
	 */
	public class LoopedSampleProcessor implements IAudioProcessor 
	{
		private var _deltaSamples : int; /** Start position within the first buffer to fill. */
		private var _pitch : int;
		private var _velocity : int;
		private var _sourceSamples : Vector.<Vector.<Number>>; /** The source samples to play. */
		private var _sourceStartIndex : int; /** Where to start playback in the source vector. */
		private var _sourceEndIndex : int; /** Where to end playback in the source vector. */
		private var _sourceIndexNum : Number; /** The index in the source vector at the start of each buffer. */
		private var _sourceIndexInt : int; /** _sourceIndexNum floored. */
		private var _sourceRelativeIndex : int = 0; /** Index relative to _sourceStartIndex instead of 0. */
		private var _bufferIndex : int; /** The write position in the buffer. */
		private var _bufferEndIndex : int; /** Last sample in buffer to write to. */
		private var _attackDuration : int = 50;
		private var _decayDuration : int = 100;
		private var _gain : Number = 1;
		private var _rate : Number;
		private var _loop : Boolean;
		private var _isStopped : Boolean;

		/**
		 * @param samples : Vector with the source sound samples.
		 * @param deltaSamples : offset into the buffer where to start writing.
		 */
		public function LoopedSampleProcessor(samples : Vector.<Vector.<Number>>, start : int, end : int, loop : Boolean, deltaSamples : int = 0, pitch : int = 60, velocity : int = 100)
		{
			_sourceSamples = samples;
			_sourceStartIndex = start;
			_sourceEndIndex = end;
			_loop = loop;
			_deltaSamples = deltaSamples;
			_pitch = pitch;
			_velocity = velocity;
			_sourceIndexNum = start + 0.001;
			_sourceIndexInt = start;
			_sourceRelativeIndex = 0;
		}

		/**
		 * @param buffer : Stereo buffer of AudioDriver.BUFFER_SIZE number of samples.
		 * @param bufferStartIndex: From where in the buffer to start to write.
		 * @param bufferLength: Number of samples to write.
		 * @return Boolean : True if playback is done and this generator can be removed.
		 */
		public function process(buffer : Vector.<Vector.<Number>>, bufferStartIndex : uint, bufferLength : uint) : Boolean
		{
			_bufferIndex = bufferStartIndex + _deltaSamples;
			_bufferEndIndex = bufferStartIndex + bufferLength;
			_deltaSamples = 0;
			
			if(_loop && !_isStopped)
			{
				/** As long as the end of the buffer isn't reached. */
				while(_bufferIndex < _bufferEndIndex)
				{
					writeSamples(buffer, _bufferEndIndex - _bufferIndex);
					
					/** End of source samples reached. Begin reading from start again. */
					if(_sourceIndexNum >= _sourceEndIndex - 4)
					{
						_sourceIndexNum = _sourceStartIndex + 0.001;
						_sourceIndexInt = _sourceStartIndex;
						_sourceRelativeIndex = 0;
					}
				}
			}
			else /** No loop. */
			{
				writeSamples(buffer, bufferLength);
			}
			
			/** Sound ends if these conditions are met. */
			if(_sourceIndexNum >= _sourceEndIndex - 2 && (_isStopped || !_loop)) return true;
			
			return false;
		}

		/**
		 * 
		 */
		private function writeSamples(buffer : Vector.<Vector.<Number>>, bufferLength : uint) : void
		{
			var bufferLeft : Vector.<Number> = buffer[0]; /** left channel */
			var bufferRight : Vector.<Number> = buffer[1]; /** right channel */
			var sourceLeft : Vector.<Number> = _sourceSamples[0]; /** left channel */
			var sourceRight : Vector.<Number> = _sourceSamples[1]; /** right channel */
			
			/** Calculate the number of source samples required to process one buffer (+2 for interpolation). */
			var sourceSamplesRequired : int = Math.ceil(bufferLength * _rate) + 2;
			
			/** If the end of the slice is reached within the current buffer, adjust the samples to read. */
			if(_sourceIndexInt + sourceSamplesRequired <= _sourceEndIndex) var sourceSamplesAvailable : int = sourceSamplesRequired;
			else sourceSamplesAvailable = _sourceEndIndex - _sourceIndexInt;
			
			/** Get the amount of samples to write. Will be less than bufferLength at the last time process is called in no-loop mode. */
			var samplesToWrite : int = (sourceSamplesAvailable == sourceSamplesRequired) ? bufferLength : (sourceSamplesAvailable - 2) / _rate;
			
			/** Get the last index in the buffer to write to. */
			var bufferEndIndex : int = _bufferIndex + samplesToWrite;
			
			while(_bufferIndex < bufferEndIndex)
			{
				/** A short attack to remove clicks at the start of the slice */
				if(_sourceRelativeIndex < _attackDuration) _gain = (_sourceRelativeIndex / _attackDuration);
				
				/** A short decay to remove clicks at the end of the slice. */
				if(_sourceEndIndex - _sourceRelativeIndex < _decayDuration) _gain = (_sourceEndIndex - _sourceRelativeIndex) / _decayDuration;
				
				/** The interpolation value between 0 and 1. */
				var alpha : Number = _sourceIndexNum - _sourceIndexInt;
				
				/** Avoid reading equal samples when _rate < 1.0. */
				if(int(alpha) != alpha)
				{
					/** Read two stereo samples (for linear interpolation). */
					var l0 : Number = sourceLeft[_sourceIndexInt];
					var r0 : Number = sourceRight[_sourceIndexInt];
					
					var l1 : Number = sourceLeft[_sourceIndexInt + 1];
					var r1 : Number = sourceRight[_sourceIndexInt + 1];
				}
				
				/** Add interpolated amplitudes to stream. */
				bufferLeft[_bufferIndex] += (l0 + alpha * (l1 - l0)) * _gain;
				bufferRight[_bufferIndex] += (r0 + alpha * (r1 - r0)) * _gain;
				
				/** Increase source and buffer positions. */
				_sourceIndexNum += _rate;
				_sourceIndexInt = int(_sourceIndexNum);
				_sourceRelativeIndex = _sourceIndexNum - _sourceStartIndex;
				_bufferIndex ++;
			}
		}

		/**
		 * 
		 */
		public function stop() : void
		{
			/** If possible leave the short decay to have a nice smooth end to the sound. */
			_sourceEndIndex = Math.min(_sourceIndexNum + _decayDuration, _sourceEndIndex);
			_isStopped = true;
		}
		
		/**
		 * 
		 */
		public function set position(position : uint) : void
		{
			_sourceIndexNum = position + 0.001;
			_sourceIndexInt = position;
			_sourceRelativeIndex = _sourceIndexNum - _sourceStartIndex;
		}

		/**
		 * 
		 */
		public function get pitch() : int
		{
			return _pitch;
		}

		/**
		 * 
		 */
		public function get sampleIndex() : int
		{
			return _sourceIndexInt;
		}

		/**
		 * 
		 */
		public function get rate() : Number
		{
			return _rate;
		}

		public function set rate(value : Number) : void
		{
			_rate = Math.max(0, value);
		}
		
		/**
		 * 
		 */
		public function set start(value : Number) : void
		{
			_sourceStartIndex = value;
		}
		
		/**
		 * 
		 */
		public function set end(value : Number) : void
		{
			_sourceEndIndex = value;
		}
	}
}
