package com.hisschemoller.audio.model.processor 
{

	/**
	 * Plays part of an MP3 file (a slice) from _startSampleIndex to _endSampleIndex.
	 * 
	 * @author Wouter Hisschemöller
	 * (c) Feb 20, 2009
	 */
	public class PitchedSampleProcessor implements IAudioProcessor 
	{
		private var _deltaSamples : int; /** Start position within the first buffer to fill. */
		private var _pitch : int;
		private var _velocity : int;
		private var _sourceSamples : Vector.<Vector.<Number>>; /** The source samples to play. */
		private var _startSampleIndex : int; /** Where to start playback in the source vector. */
		private var _endSampleIndex : int; /** Where to end playback in the source vector. */
		private var _currentSampleIndexNum : Number; /** The index in the source vector at the start of each buffer. */
		private var _currentSampleIndexInt : int; /** _currentSampleIndexNum floored. */
		private var _relativeSampleIndex : int = 0; /** Convenient index relative to _startSampleIndex instead of 0. */
		private var _numSamplesRequired : int;
		private var _attackDuration : int = 50;
		private var _decayDuration : int = 100;
		private var _gain : Number = 1;
		private var _rate : Number;

		/**
		 * @param samples : Vector with the source sound samples.
		 * @param deltaSamples : offset into the buffer where to start writing.
		 */
		public function PitchedSampleProcessor (samples : Vector.<Vector.<Number>>, start : int, end : int, deltaSamples : int = 0, pitch : int = 60, velocity : int = 100)
		{
			_sourceSamples = samples;
			_startSampleIndex = start;
			_endSampleIndex = end;
			_deltaSamples = deltaSamples;
			_pitch = pitch;
			_velocity = velocity;
			_currentSampleIndexNum = _startSampleIndex;
		}

		/**
		 * @param buffer : stereo buffer of AudioDriver.BUFFER_SIZE number of samples
		 * @param bufferStartIndex: From where in the buffer to start to write.
		 * @param bufferLength: Number of samples to write.
		 * @return Boolean : true if the playback is done and this generator can be removed
		 */
		public function process (buffer : Vector.<Vector.<Number>>, bufferStartIndex : uint, bufferLength : uint) : Boolean
		{
			try
			{
				var bufferLeft : Vector.<Number> = buffer[0]; /** left channel */
				var bufferRight : Vector.<Number> = buffer[1]; /** right channel */
				var sourceLeft : Vector.<Number> = _sourceSamples[0]; /** left channel */
				var sourceRight : Vector.<Number> = _sourceSamples[1]; /** right channel */

				var scaledBufferSize : Number = bufferLength * _rate;
				_currentSampleIndexInt = _currentSampleIndexNum;
				var alpha : Number = _currentSampleIndexNum - _currentSampleIndexInt;
				
				var positionTargetNum : Number = alpha;
				var positionTargetInt : int = -1;
	
				/** Compute the number of samples needed to process one buffer (+2 for interpolation). */
				_numSamplesRequired = Math.ceil( scaledBufferSize ) + 2;
				
				/** If the end of the slice is reached within the current buffer, adjust the samples to read. */
				if(_currentSampleIndexInt + _numSamplesRequired <= _endSampleIndex)
				{
					var numSamplesAvailable : int = _numSamplesRequired;
				}
				else
				{
					numSamplesAvailable = _endSampleIndex - _currentSampleIndexInt;
				}
				
				/** Get the amount of samples to write. Will be less than bufferLength at the last time process is called. */
				var n : int = (numSamplesAvailable == _numSamplesRequired) ? bufferLength : (numSamplesAvailable - 2) / _rate;
				
				for(var i : int = _deltaSamples; i < n; ++i)
				{
					/** A short attack to remove clicks at the start of the slice */
					if(_relativeSampleIndex + i < _attackDuration)
					{
						_gain = ((_relativeSampleIndex + i) / _attackDuration);
					}
					
					/** A short decay to remove clicks at the end of the slice. */
					if(_endSampleIndex - _relativeSampleIndex + i < _decayDuration)
					{ 
						_gain = (_endSampleIndex - _relativeSampleIndex + i) / _decayDuration;
					}
					
					/** Avoid reading equal samples if _rate < 1.0. */
					if(int( positionTargetNum ) != positionTargetInt)
					{
						positionTargetInt = positionTargetNum;
						
						/** Set vector read position. */
						var position : int = _currentSampleIndexInt + positionTargetInt;
						
						/** Read two stereo samples (for linear interpolation). */
						var l0 : Number = sourceLeft[position];
						var r0 : Number = sourceRight[position];
						
						var l1 : Number = sourceLeft[position + 1];
						var r1 : Number = sourceRight[position + 1];
					}
					
					/** Add interpolated amplitudes to stream. */
					var bufferPosition : int = i + bufferStartIndex;
					bufferLeft[bufferPosition] += (l0 + alpha * (l1 - l0)) * _gain;
					bufferRight[bufferPosition] += (r0 + alpha * (r1 - r0)) * _gain;
					
					/** Increase target position. */
					positionTargetNum += _rate;
					
					/** Increase fraction and clamp between 0 and 1. */
					alpha += _rate;
					while(alpha >= 1.0) --alpha;
				}
				
				/** If the slice is done, fill the rest of the buffer with silence. */
				if(i < bufferLength)
				{
					while(i < bufferLength)
					{
						bufferPosition = i + bufferStartIndex;
						bufferLeft[bufferPosition] += 0.0;
						bufferRight[bufferPosition] += 0.0;
						
						++i;
					}
				}
				
				_currentSampleIndexNum += scaledBufferSize - _deltaSamples;
				_relativeSampleIndex = _currentSampleIndexNum - _startSampleIndex;
				_deltaSamples = 0;
				
				if(_currentSampleIndexNum >= _endSampleIndex)
				{
					return true;
				}
				
				return false;
			}
			catch (error : Error)
			{
				trace( "Error: " + error );
				trace( "_deltaSamples: " + _deltaSamples );
				trace( "i: " + i );
				trace( "bufferStartIndex: " + bufferStartIndex );
				trace( "bufferPosition: " + bufferPosition );
				return true;
			}
			
			return false;
		}

		/**
		 * 
		 */
		public function stop () : void
		{
			/** If possible leave the short decay to have a nice smooth end to the sound. */
			_endSampleIndex = Math.min( _currentSampleIndexNum + _decayDuration, _endSampleIndex );
		}

		/**
		 * 
		 */
		public function get pitch () : int
		{
			return _pitch;
		}

		/**
		 * 
		 */
		public function get sampleIndex () : int
		{
			return _currentSampleIndexInt;
		}

		/**
		 * 
		 */
		public function get rate () : Number
		{
			return _rate;
		}

		public function set rate (value : Number) : void
		{
			if(value < 0.0) value = 0;

			_rate = value;
		}
	}
}
